/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Objects;
import java.util.RandomAccess;
import java.util.function.Consumer;

/**
 * <p>Dynamic, mutable, wrapper for Java Collection Framework's List.</p>
 *
 * <p>This dynamic object array is constructed by the static {@code newDynamicArray} methods of {@code Arrays}.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Dynamic arrays, such as this class, always report {@code DYNAMIC} and {@code MUTABLE}.</p>
 *
 * <p>This class attempts to establish these characteristics to accommodate these optimizations at best effort. This
 * means that the array implementation determines which set of characteristics apply to its elements without traversing
 * these elements, in other words, without incurring a performance penalty. Characteristics are always invalidated if
 * conditions of these characteristics are broken as result of array modification. Characteristics are only added in
 * scenarios in which their conditions can be 'tested' without traversal of elements. For example, the {@code SORTED}
 * characteristic is added if the length of an array is reduced to zero or one element, or if all elements are replaced
 * by an array that is marked as being sorted. The semantics of characteristics processing is documented for each
 * "destructive" method. Users can take full control of characteristics assignment by means of the
 * {@link MutableArray#doAction(ArrayAction)} method.</p>
 *
 * @author Leon van Zantvoort
 */
@NotThreadSafe
class DynamicListArrayImpl<E> extends MutableListArrayImpl<E> implements DynamicObjectArray<E> {

    /** Counter for fail-fast error detection. */
    protected transient int modCount = 0;

    /**
     * Package visible constructor used to wrap specified {@code array}.
     *
     * @param list list to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @param characteristics array characteristics.
     */
    DynamicListArrayImpl(@Nonnull List<E> list, int offset, int length, int characteristics, Comparator<? super E> comparator) {
        super(list, offset, length, characteristics, comparator);
    }

    /**
     * A version of rangeCheck used by add and addAll.
     *
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    private void rangeCheckForAdd(int index) {
        Arrays.rangeCheckForAdd(length, index);
    }

    /**
     * Fail-fast error detection. Sub classes (such as sub arrays) may override this method if they wish to implement
     * concurrent modification checking.
     *
     * @throws ConcurrentModificationException if parent array has been structurally modified.
     */
    protected void checkForComodification() {
    }

    /**
     * Updates length of this array and increases modCount.
     *
     * <p>Sub arrays must override this method and delegate the invocation this implementation by calling
     * {@code super.updateLength()}, next they must delegate the invocation to the parent array.</p>
     *
     * @param diff number of elements that are added / removed.
     */
    protected void updateLength(int diff) {
        modCount++;
        length += diff;
    }

    /**
     * Returns the length of the root array. If {@code this} array is the root array, this method returns the same
     * value is {@code length()}.
     *
     * <p>Sub arrays must override this method and delegate the invocation to the parent.</p>
     *
     * @return length of the root array.
     */
    protected int rootLength() {
        return length;
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    @Override
    protected int filterCharacteristics(int characteristics) {
        return characteristics | MUTABLE | DYNAMIC;
    }

    /**
     * This method has no use because the backing data store is a list.
     *
     * @param flag you might as well pass it Random.nextBoolean().
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void setClearReservedElements(boolean flag) {
    }

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within this array.
     *
     * @param consumer consumer that is called for each value within this array.
     */
    @SuppressWarnings("ConstantConditions")
    @Override
    public void forEach(@Nonnull Consumer<? super E> consumer) {
        Objects.requireNonNull(consumer);
        int expectedModCount = modCount;
        Iterator<E> it = elements.listIterator(offset);
        for (int index = 0; modCount == expectedModCount && index < length; index++) {
            consumer.accept(it.next());
        }
        if (modCount != expectedModCount) {
            throw new ConcurrentModificationException();
        }
    }

    /**
     * Appends the specified object to the end of this array's range. Length of {@code array} is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code object} is not {@code null};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array. Arrays
     *   sorted according to natural order are not regarded sorted if they contain {@code null} element(s);</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param object object to to be added.
     * @throws NullPointerException if object is null and array does not permit null elements.
     */
    @Override
    public void add(E object) {
        int elementIndex = offset + length;
        elements.add(elementIndex, object);
        updateLength(1);
        if (object == null) {
            removeCharacteristics(NONNULL);
        }
        if (hasCharacteristics(SORTED)) {
            updateCharacteristics(elementIndex, object);
        }
    }

    /**
     * Inserts the specified object at the specified {@code index}, moving elements at and after {@code index} to the
     * right. Length of {@code this} array is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code object} is not {@code null};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array. Arrays
     *   sorted according to natural order are not regarded sorted if they contain {@code null} element(s);</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index insertion index, relative to the offset of the array.
     * @param object object to be inserted.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if object is null and array does not permit null elements.
     */
    @Override
    public void add(int index, E object) {
        rangeCheckForAdd(index);
        int elementIndex = offset + index;
        elements.add(elementIndex, object);
        updateLength(1);
        if (object == null) {
            removeCharacteristics(NONNULL);
        }
        if (hasCharacteristics(SORTED)) {
            updateCharacteristics(elementIndex, object);
        }
    }

    /**
     * Appends the elements of the specified array to the end of this array. Length of {@code this} array is increased
     * by {@code array.length()}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param array array to be added.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     */
    @Override
    public void addAll(@Nonnull Array<? extends E> array) {
        int arrayLength = array.length();
        if (arrayLength > 0) {
            int elementIndex = offset + length;
            elements.addAll(elementIndex, ArrayBridge.asList(array));
            updateLength(arrayLength);
            retainCharacteristics(array.characteristics());
            if (hasCharacteristics(SORTED)) {
                updateCharacteristics(elementIndex, array);
            }
        }
    }

    /**
     * Inserts the elements of the specified array at the specified {@code index}, moving elements at and after
     * {@code index} to the right. Length of {@code this} array is increased by {@code array.length()}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param index insertion index, relative to the offset of the array.
     * @param array array to be inserted.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     */
    @Override
    public void addAll(int index, @Nonnull Array<? extends E> array) {
        rangeCheckForAdd(index);
        int arrayLength = array.length();
        if (arrayLength > 0) {
            int elementIndex = offset + index;
            elements.addAll(elementIndex, ArrayBridge.asList(array));
            updateLength(arrayLength);
            retainCharacteristics(array.characteristics());
            if (hasCharacteristics(SORTED)) {
                updateCharacteristics(elementIndex, array);
            }
        }
    }

    /**
     * Replaces all elements of this array by the elements of the specified array. The semantics of this method are less
     * strict than the ones expressed by {@code MutableArray}; The array that is passed as argument does not need to
     * have the same length as {@code this} array. This method does not throw an {@code IllegalArgumentException}.
     *
     * <p>The length of {@code this} array is increased or decreased by length difference between {@code this} array and
     * {@code array} argument.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param array array to replace elements of this array.
     * @throws NullPointerException if the specified array is null or if the element is replaced with a null value
     * and this array does not permit null elements.
     */
    @Override
    public void setAll(@Nonnull Array<? extends E> array) {
        int lengthDiff = array.length() - length;
        elements.subList(offset, offset + length).clear();
        elements.addAll(offset, ArrayBridge.asList(array));
        if (lengthDiff != 0) {
            updateLength(lengthDiff);
        }
        characteristics(array.characteristics());
        if (hasCharacteristics(SORTED)) {
            updateCharacteristics(offset, array);
        }
    }

    /**
     * Removes element at specified {@code index}, moving all elements after {@code index} to the left. Length of
     * {@code this} array is decreased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains no elements after this invocation, {@code NONNULL}, {@code SORTED} and
     *   {@code DISTINCT} are reported.
     * </ul>
     *
     * @param index removal index, relative to the offset of the array.
     * @return element that is removed.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @SuppressWarnings("unchecked")
    @Override
    public E remove(int index) {
        rangeCheck(index);
        E oldValue = elements.remove(offset + index);
        updateLength(-1);
        addCharacteristics(length == 0 ? NONNULL | SORTED | DISTINCT : 0);
        return oldValue;
    }

    /**
     * Removes the first occurrence of the specified object from this array if it is present. If this array does not
     * contain the object, it is unchanged. This method supports boxed primitives. This means that if this array
     * represents a primitive array, this method removes the first primitive element that is the unboxed counterpart of
     * the object.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics to lookup the element to be removed, otherwise all elements are traversed until the specified
     * object is found.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains no elements after this invocation, {@code NONNULL}, {@code SORTED} and
     *   {@code DISTINCT} are reported.
     * </ul>
     *
     * @param object object to be removed.
     * @return {@code true} if this array contained the specified object.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if this array is sorted and comparator not compatible with object.
     */
    @SuppressWarnings("unchecked")
    @Override
    public boolean remove(Object object) {
        if (hasCharacteristics(SORTED) && elements instanceof RandomAccess) {
            // Optimization only possible if list supports fast random access.
            Comparator<? super E> comparator = getComparator();
            if (object == null && comparator == null) { // No NullPointerException in case of natural ordering.
                return false;
            }
            int index = Arrays.binarySearch(elements, offset, length, (E) object, comparator);
            if (index >= 0) {
                remove(index);
                return true;
            }
            return false;
        }
        int index = indexOf(object);
        if (index >= 0) {
            remove(index);
            return true;
        }
        return false;
    }

    /**
     * Removes all of this array's elements that are also contained in the specified array. Primitives and their object
     * counterparts are regarded equal in this respect. After this call returns, this array will contain no elements
     * in common with the specified array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains no elements after this invocation, {@code NONNULL}, {@code SORTED} and
     *   {@code DISTINCT} are reported.
     * </ul>
     *
     * @param array array containing elements to be removed from this array.
     * @return {@code true} if this array changed as a result of the call.
     * @throws NullPointerException if the specified array is null. Exception may also be thrown if this or specified
     * array contain null elements.
     * @throws ClassCastException if this or specified array is sorted and an element of this array is incompatible with
     * the specified array.
     */
    @Override
    public boolean removeAll(@Nonnull Array<?> array) {
        if (hasCharacteristics(DISTINCT) && length > array.length()) {    // Optimized scenario.
            boolean modified = false;
            for (Object key : array) {
                modified |= remove(key);
            }
            return modified;
        }
        List<E> subList = elements.subList(offset, offset + length);
        subList.removeAll(ArrayBridge.asList(array));
        int removed = length - subList.size();
        assert removed >= 0;
        updateLength(-removed);
        addCharacteristics(length == 0 ? NONNULL | SORTED | DISTINCT : 0);
        return removed > 0;
    }

    /**
     * Retains only the elements in this array that are contained in the specified array. In other words, removes from
     * this array all of its elements that are not contained in the specified array. Primitives and their object
     * counterparts are regarded equal in this respect.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains no elements after this invocation, {@code NONNULL}, {@code SORTED} and
     *   {@code DISTINCT} are reported.
     * </ul>
     *
     * @param array array containing elements to be retained in this array.
     * @return {@code true} if this array changed as a result of the call.
     * @throws NullPointerException if the specified array is null. Exception may also be thrown if this or specified
     * array contain null elements.
     * @throws ClassCastException if this or specified array is sorted and an element of this array is incompatible with
     * the specified array.
     */
    @Override
    public boolean retainAll(@Nonnull Array<?> array) {
        List<E> subList = elements.subList(offset, offset + length);
        subList.retainAll(ArrayBridge.asList(array));
        int removed = length - subList.size();
        assert removed >= 0;
        updateLength(-removed);
        addCharacteristics(length == 0 ? NONNULL | SORTED | DISTINCT : 0);
        return removed > 0;
    }

    /**
     * Removes all elements of range, moving elements from the right after range to offset position.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>After this invocation, {@code NONNULL}, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     */
    @Override
    public void clear() {
        setAll(ObjectArray.<E>emptyArray());
    }

    /**
     * This method has no use when the backing data store is a list.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void compact() {
    }

    /**
     * Removes all duplicates from this array. The order of elements may be altered as result of this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} and {@code DISTINCT} are reported after this invocation.</li>
     * </ul>
     *
     * @throws NullPointerException if this array contains null elements.
     * @throws ClassCastException if this array contains elements that are not mutually {@code Comparable}.
     */
    @Override
    public void distinct() {
        sort();
        fastDistinct();
    }

    /**
     * Removes all duplicates from this array. The order of elements may be altered as result of this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} and {@code DISTINCT} are reported after this invocation.</li>
     * </ul>
     *
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> should be used
     */
    @Override
    public void distinct(Comparator<? super E> comparator) {
        sort(comparator);
        fastDistinct();
    }

    /**
     * Optimized distinct implementation.
     */
    private void fastDistinct() {
        if (!hasCharacteristics(DISTINCT)) {
            int removed = 0;
            ListIterator<E> it = elements.listIterator(offset);
            if (it.hasNext()) {
                final int fromIndex = offset;
                final int toIndex = fromIndex + length;
                E previous = it.next();
                for (int index = fromIndex + 1; index < toIndex; index++) {
                    E next = it.next();
                    if (Objects.equals(next, previous)) {
                        it.remove();
                        removed++;
                    }
                    previous = next;
                }
            }
            updateLength(-removed);
            addCharacteristics(DISTINCT);
        }
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SubDynamicListArrayImpl<E> extends DynamicListArrayImpl<E> {

        final DynamicListArrayImpl<E> parent;

        @SuppressWarnings("unchecked")
        SubDynamicListArrayImpl(@Nonnull DynamicListArrayImpl<E> parent, int offset, int length) {
            super(parent.elements, offset, length, parent.characteristics, parent.comparator);
            this.parent = parent;
            this.modCount = parent.modCount;
        }

        @Override
        protected void checkForComodification() {
            if (modCount != parent.modCount) {
                throw new ConcurrentModificationException();
            }
            parent.checkForComodification();
        }

        @Override
        protected int rootLength() {
            return parent.rootLength();
        }

        @Override
        protected void updateLength(int diff) {
            super.updateLength(diff);
            parent.updateLength(diff);
        }

        @Override
        protected void updateCharacteristics(int elementIndex, E object) {
            super.updateCharacteristics(elementIndex, object);
            if (parent.hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, object);
            }
        }

        @Override
        protected void updateCharacteristics(int elementIndex, @Nonnull Array<? extends E> array) {
            super.updateCharacteristics(elementIndex, array);
            if (parent.hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, array);
            }
        }

        @Override
        protected void setComparator(@Nullable Comparator<? super E> comparator) {
            // Comparator change of sub array means that sorted status of parent array must be reset.
            super.setComparator(comparator);
            parent.removeCharacteristics(SORTED);
            parent.setComparator(null);
        }

        @Override
        public int characteristics(int characteristics) {
            // Parent array retains characteristics that are shared with sub array. Additional characteristics
            // assigned to sub array are not set for parent array, as it is not guaranteed that these
            // characteristics hold for parent array.
            parent.retainCharacteristics(characteristics);
            return super.characteristics(characteristics);
        }

        @Override
        public ArrayIterator<E> iterator() {
            checkForComodification();
            return super.iterator();
        }

        @Override
        public ArrayIterator<E> iterator(int index) {
            checkForComodification();
            return super.iterator(index);
        }

        @Override
        public void forEach(@Nonnull Consumer<? super E> consumer) {
            checkForComodification();
            super.forEach(consumer);
        }

        @Override
        public E get(int index) {
            checkForComodification();
            return super.get(index);
        }

        @Override
        public @Nonnull E getNonnull(int index) {
             checkForComodification();
             return super.getNonnull(index);
        }

        @Override
        public E set(int index, E value) {
            checkForComodification();
            return super.set(index, value);
        }

        @Override
        public void setAll(@Nonnull Array<? extends E> array) {
            checkForComodification();
            super.setAll(array);
        }

        @Override
        public void add(E object) {
            checkForComodification();
            super.add(object);
        }

        @Override
        public void add(int index, E object) {
            checkForComodification();
            super.add(index, object);
        }

        @Override
        public void addAll(@Nonnull Array<? extends E> array) {
            checkForComodification();
            super.addAll(array);
        }

        @Override
        public void addAll(int index, @Nonnull Array<? extends E> array) {
            checkForComodification();
            super.addAll(index, array);
        }

        @Override
        public boolean remove(Object object) {
            checkForComodification();
            return super.remove(object);
        }

        @Override
        public E remove(int index) {
            checkForComodification();
            return super.remove(index);
        }

        @Override
        public boolean removeAll(@Nonnull Array<?> array) {
            checkForComodification();
            return super.removeAll(array);
        }

        @Override
        public boolean retainAll(@Nonnull Array<?> array) {
            checkForComodification();
            return super.retainAll(array);
        }

        @Override
        public Object[] toArray() {
            checkForComodification();
            return super.toArray();
        }

        @Override
        public <T> T[] toArray(@Nonnull T[] array) {
            checkForComodification();
            return super.toArray(array);
        }
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public DynamicObjectArray<E> offset(int offset) {
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new SubDynamicListArrayImpl<>(this, newOffset, newLength);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    public DynamicObjectArray<E> length(int length) {
        subArrayCheck(offset, length);
        return new SubDynamicListArrayImpl<>(this, offset, length);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    public DynamicObjectArray<E> subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        subArrayCheck(newOffset, newLength);
        return new SubDynamicListArrayImpl<>(this, newOffset, newLength);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator<E> {
        final ListIterator<E> iterator;

        /**
         * The modCount value that the iterator believes that the backing
         * array should have. If this expectation is violated, the iterator
         * has detected concurrent modification.
         */
        int expectedModCount = modCount;

        ArrayIteratorImpl(ListIterator<E> iterator) {
            this.iterator = iterator;
        }

        void checkForComodification() {
            if (modCount != expectedModCount) {
                throw new ConcurrentModificationException();
            }
        }

        @Override
        public boolean hasNext() {
            return iterator.hasNext();
        }

        @Override
        public E next() {
            checkForComodification();
            return iterator.next();
        }

        @Override
        public void remove() {
            checkForComodification();
            iterator.remove();
            updateLength(-1);
            if (length <= 1) {
                addCharacteristics(SORTED | DISTINCT);
            }
            expectedModCount = modCount;
        }

        @Override
        public boolean hasPrevious() {
            return iterator.hasPrevious();
        }

        @Override
        public E previous() {
            checkForComodification();
            return iterator.previous();
        }

        @Override
        public int nextIndex() {
            return iterator.nextIndex();
        }

        @Override
        public int previousIndex() {
            return iterator.previousIndex();
        }

        @SuppressWarnings("unchecked")
        @Override
        public void set(E object) {
            checkForComodification();
            iterator.set(object);
            if (length > 1) {      // Unable to compare against replaced object.
                removeCharacteristics(SORTED | DISTINCT);
            }
        }

        @Override
        public void add(E object) {
            checkForComodification();
            iterator.add(object);
            updateLength(1);
            if (length > 1) {
                removeCharacteristics(SORTED | DISTINCT);
            }
            expectedModCount = modCount;
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator<E> iterator() {
        return new ArrayIteratorImpl(elements.subList(offset, offset + length).listIterator());
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator<E> iterator(int index) {
        if (index > length) {
            throw new IndexOutOfBoundsException();
        }
        return new ArrayIteratorImpl(elements.subList(offset, offset + length).listIterator(index));
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    @Override
    Object writeReplace() {
        return new SerializationProxy<>(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy<E> implements Serializable {
        static final long serialVersionUID = -8436338728620340331L;

        transient int offset;
        final int length;
        final int characteristics;
        final Comparator<? super E> comparator;
        transient List<E> elements;

        SerializationProxy(DynamicListArrayImpl<E> array) {
            this.offset = array.offset;
            this.length = array.length;
            this.characteristics = array.characteristics;
            this.comparator = array.comparator;
            this.elements = array.elements;
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            s.writeObject(new ArrayList<>(elements.subList(offset, offset + length)));
        }

        @SuppressWarnings("unchecked")
        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = (List<E>) s.readObject();
        }

        @SuppressWarnings("unchecked")
        private Object readResolve() {
            return new DynamicListArrayImpl<>(elements, 0, length, characteristics, comparator);
        }
    }
}
